from functools import wraps
import cls


class Base(metaclass=cls.ClsMeta):
    @cls
    def decor1(cls, decor_arg):
        def wrap(func):
            @wraps(func)
            def wrapper(self, arg):
                pre = 'cdp({})'.format(cls.__name__)
                retval = func(self, arg)
                sub = 'cds({})'.format(decor_arg)
                return '.'.join((pre, retval, sub))
            return wrapper
        return wrap

    @cls(True)
    def decor2(cls):
        def wrap(func):
            @wraps(func)
            def wrapper(self, arg):
                pre = 'cdp2({})'.format(cls.__name__)
                retval = func(self, arg)
                sub = 'cds2()'.format()
                return '|'.join((pre, retval, sub))
            return wrapper
        return wrap


def plain_decor(func):
    pre = 'pdb'
    sub = 'pds'
    return lambda x: '|'.join((pre, func(x,), sub))


class Extended(Base):
    @cls
    def decor3(cls, decor_arg='decor3 arg'):
        def wrap(func):
            @wraps(func)
            def wrapper(self, arg):
                pre = 'cdp3({})'.format(cls.__name__)
                retval = func(self, arg)
                sub = 'cds3({})'.format(decor_arg)
                return '|'.join((pre, retval, sub))
            return wrapper
        return wrap

    @cls.decor3()
    @cls.decor1('decor1 arg')
    @cls.decor2
    def func(self, func_arg):
        return func_arg


def test_cls():
    obj = Extended()
    retval = obj.func('function arg')
    print(retval)
    assert retval == 'cdp3(Extended)|cdp(Extended).cdp2(Extended)|function arg|cds2().cds(decor1 arg)|cds3(decor3 arg)'


if __name__ == '__main__':
    test_cls()



from functools import wraps


class Stub():
    def __init__(self, func):
        self.func = func
        self.level = 0

    def __call__(self, *args, **kwargs):
        raise RuntimeError('Stubs are not to be called!')

    @classmethod
    def wraps(cls, func):
        if isinstance(func, cls):
            func.level += 1
            return func
        return wraps(func)(cls(func))


class Wrapper():
    def __init__(self, func, no_arg=False):
        assert callable(func), '{}'.format(func)
        self.func = func
        self.no_arg = no_arg
        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class DecoratorDelegator():
    def __init__(self, namespace):
        self.namespace = namespace
        self.todo = []

    def __call__(self, decorator):
        if isinstance(decorator, bool) and decorator:
            # if take no (other) arg
            def func(decorator):
                return Wrapper(decorator, True)
            return func
        if not callable(decorator):
            msg = ('@cls applies to methods. If you have more decorators, for '
                   'example, @property and @classmethod, please consider '
                   'different order.')
            raise RuntimeError(msg)
        return Wrapper(decorator)

    def __getattr__(self, decorator_name):
        def collect_args(*args, **kwargs):
            def delay(func):
                stub = Stub.wraps(func)
                self.todo.append((decorator_name, args, kwargs, stub))
                return stub
            if len(args) == 1 and callable(args[0]):
                return delay(args[0])
            return delay
        return collect_args

    def realize(self, cls_):
        for decorator_name, args, kwargs, stub in self.todo:
            decorator = getattr(cls_, decorator_name)
            if decorator.no_arg:
                wrapped = decorator(cls_)(stub.func)
            else:
                wrapped = decorator(cls_, *args, **kwargs)(stub.func)
            if stub.level > 0:
                stub.level -= 1
                stub.func = wrapped
            else:
                self.namespace[stub.__name__] = wrapped
                setattr(cls_, stub.__name__, wrapped)


# Steps constructing class:
# 0. Resolving MRO entries
# 1. Determining the appropriate metaclass
class ClsMeta(type):
    # 2. Preparing the class namespace
    def __prepare__(name, bases, **kwds):
        namespace = type.__prepare__(name, bases, **kwds)
        namespace['cls'] = DecoratorDelegator(namespace)
        return namespace

    # 3. Executing the class body
    # 4. Creating the class object
    # def __call__(name, bases, namespace, **kwds):
    #     def __new__(meta, name, bases, namespace, **kwds):
        #     4.1.0 colloct descriptors
        #     4.1.1 for each descriptor call __set_name__(self, owner, name)
        #     4.1.2 init cls
    def __init__(cls, name, bases, namespace):
        delegator = namespace.pop('cls')
        delegator.realize(cls)
        cls = type.__init__(cls, name, bases, namespace)
        return cls

    # 5. Nothing more, ready to go



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="acls",
    version="1.0.2",
    author="guoquan",
    author_email="guoquanscu@gmail.com",
    description="`@cls` - Class Made Aware to Decorator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/guoquan/acls",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import sys
from .base import ClsMeta


def __getattr__(_):
    pass


class Quiet(sys.modules[__name__].__class__):
    def __call__(self, *_, **__):
        pass


sys.modules[__name__].__class__ = Quiet
